#!/bin/bash
#
# Licensed Materials - Property of IBM
#
# IBM Financial Crimes Insight with Watson PID 5900-A0H
# IBM Financial Crimes Insight with Watson, Private PID 5737-E41
#
# (C) Copyright IBM Corp. 2018  All rights reserved.
#
# The source code for this program is not published or otherwise divested
# of its trade secrets, irrespective of what has been deposited with the
# U. S. Copyright Office.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# DISCLAIMER OF WARRANTIES.  The following code is sample code
# created by IBM Corporation.  This sample code is part of
# IBM Financial Crimes Insight with Watson and is warranted to perform its
# intended function only if used unmodified.  If you modify
# this code then it is considered provided "AS IS", without
# warranty of any kind.  Notwithstanding the foregoing, IBM
# shall not be liable for any damages arising out of your use
# of the sample code, even if they have been advised of the
# possibility of such damages.
#

# set global variables
function set_vars {
  # Colors
  B='\033[0;31m' # bold
  N='\033[0m' # normal

  # this script
  SCRIPT_DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" >/dev/null && pwd )"

  # words not to save into logs -- i.e., grep -v $STOP_WORDS
  STOP_WORDS=("" "pass" "passwd" "password" "xor" "db2inst1" "elasticuser" \
           "fci_kafka_msg_jks_aliasname" "fciauthid" "jwt_key" "mongodb_admin_user" \
           "mongodb_iui_user" "mongodb_root" "mongodb_rs_key" "mqm:" "credential")
  STOP_OPTS=$(echo "${STOP_WORDS[@]}" | sed 's/ / -e /g')

  # create temp directory
  t=$(mktemp -d)
  [[ $? -ne 0 ]] && { echo "Error: could not set temp directory."; exit; }
  [[ -z "${t}" ]] && { echo "Error: temp directory path is unexpectedly empty!?"; exit; }
  [[ -d "${t}" ]] || { echo "Error: the temp directory is not a directory!?"; exit; }
  chmod 700 "${t}" # tighten permissions
}

# display welcome and instructions
function get_welcome {
  clear
  echo -e "Welcome to fcisupport.sh\n"
  echo "This script gathers troubleshooting information about your IBM Financial Crimes Insight (FCI) installation."
  echo "It is designed to run on the Kubernetes Master. Please use an account that can run kubectl, docker, and helm commands."
}

# get basic information about the installation
function get_k8s_basics {
  printf "\nGathering information. Please wait up to 15 minutes...\n"

  while read -r fn cmd; do eval "${cmd}" &> "$t/${fn}.log"; done <<'EOF'
   readme echo "This tarball was created on the Kubernetes Master and contains various troubleshooting details which customers may remove or redact as needed."
   osinfo cat /etc/redhat-release; cat /etc/os-release
   hostname-f hostname -f
   hostname-I hostname -I
   etc-hosts cat /etc/hosts
   docker-version docker -v
   kubernetes-version kubectl version
   helm-version helm version
   APARs-installed kubectl get pods --all-namespaces -o jsonpath="{..image}" | tr -s '[[:space:]]' '\n' | sort -u
   docker-stats docker stats --no-stream
   linux-kernel-version uname -r
   firewall-cmd firewall-cmd --list-all-zones
   iptables iptables -L
   systemctl-no-pager systemctl --no-pager
   ulimit ulimit -a
   df-h df -h
   free-memory-and-swap free -m
   journalctl-since-yesterday journalctl --no-pager --since yesterday
   vmstat-cpu-10-seconds vmstat 1 10 -t
   vmstat-disk-10-seconds vmstat -d -S M 1 10 -t
   var-log-messages cat /var/log/messages
   kubectl-get-all-wide kubectl get all -o wide
   kubectl-get-events kubectl get events --all-namespaces
   helm-ls helm ls
EOF
}

# get the installation logs across all installed verticals
function get_install_logs {
  # back out two directories from this script (e.g., /root)
  pushd "${SCRIPT_DIR}/../../" &>/dev/null

  # find subpaths named fci containing *.log and tar them to retain their paths and thus the vertical name
  { find . -ipath '*install-toolkit*' -iname '*.log' -print0 | tar -czvf "${t}/install-logs.tgz" --null -T - ; } &>/dev/null

  popd &>/dev/null
}

# get 'describe pod' and 'kubectl logs' for all pods and their containers in the default namespace
function get_k8s_pods_conts {
  # enumerate all pods and their containers, then get their describe and logs output
  kubectl get pods -o=jsonpath='{range .items[*]}{"\n"}{.metadata.name}{" "}{range .spec.containers[*]}{.name}{" "}{end}{end}' | sort > "${t}/pc.txt"

  sed -i '/^\s*$/d' "${t}/pc.txt" # remove any blank lines
  while read -r -a items
  do
    # remember the pod name and describe it
    pod="${items[0]}"
    kubectl describe pod "${pod}" &> "${t}/kubectl-describe-pod-${pod}"

    # remove pod from the array
    unset items[0]

    # iterate through each container
    for container in $(echo "${items[@]}"); do \

      # do "kubectl logs" for each container but remove PASSWORD= which may contain passwords
      kubectl logs $pod -c "$container" 2>&1 |\
        eval "grep -i -v -e 'PASSWORD=' ${STOP_OPTS}" &> "${t}/kubectl-logs-${pod}-c-${container}"

    done
  done < "${t}/pc.txt"

  # remove pc.txt
  \rm -f "${t}/pc.txt"
}

# run 'helm get' on all installed charts
function get_charts {
  charts=$(helm ls | awk 'NR>1 {print $1}') # enumerate installed charts

  # store each chart into an items array
  read -r -a items <<< $(echo "$charts")

  # iterate through each helm chart
  for item in "${items[@]}"; do

    # "helm get" each chart. That returns passwords though, so remove them with grep -v.
    helm get $item 2>&1 | eval "grep -i -v ${STOP_OPTS}" &> "${t}/helm-get-chart-${item}"

  done
}

# collect logs from db2
function get_db2diag {
   kubectl exec -ti $(kubectl get pods | awk '/cedm-datastore/{print $1}') -c cedm-data-store \
   -- sudo su - db2inst1 bash -c "db2diag -g level=Error" &> "${t}/cedm-datastore-db2diag-level-error"

   kubectl exec -ti $(kubectl get pods | awk '/cedm-datastore/{print $1}') -c cedm-data-store \
   -- sudo su - db2inst1 bash -c "db2diag -g level=Severe" &> "${t}/cedm-datastore-db2diag-level-severe"
}

# despite what helm says, get the actual running configmaps, in case somebody did "kubectl edit"
function get_configmaps {
   kubectl get configmap -o json | eval "grep -i -v ${STOP_OPTS}" &> "${t}/kubectl-get-configmaps.log"
}

# create a tarbar of all accumulated information
function get_tar_gz {
  output="$HOME/fcisupport.tgz"

  tar -zcvf "${output}" -C $t . 1>/dev/null
}

# clean up and say goodbye
function get_goodbye {
  [[ -z "${t}" ]] && { echo "Error: Temp directory path is empty prior to deletion!?"; exit; }
  subdircount=$(find "${t}" -maxdepth 1 -type d | wc -l)
  [[ $subdircount -eq 1 ]] || { echo "Error: There should be no subfolders in the temp directory. Please manually delete ${t}."; exit; }

  # this can be dangerous if it is empty or incorrect, thus the previous checks
  \rm -rf "${t}"

  printf "\nDone. Please upload ${B}%s${N} to IBM. \n" "$(readlink -f "${output}")"
}

# run
function main {
  get_welcome
  set_vars
  get_k8s_basics
  get_install_logs
  get_k8s_pods_conts
  get_charts
  get_db2diag
  get_configmaps
  get_tar_gz
  get_goodbye
}

# start here and pass any parameters
main "$@"
