#!/usr/bin/env bash
#
# Licensed Materials - Property of IBM
#
# IBM Financial Crimes Insight with Watson PID 5900-A0H
# IBM Financial Crimes Insight with Watson, Private PID 5737-E41
#
# (C) Copyright IBM Corp. 2017, 2018  All rights reserved.
#
# The source code for this program is not published or otherwise divested
# of its trade secrets, irrespective of what has been deposited with the
# U. S. Copyright Office.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

export TEXTDOMAIN=install.sh
. gettext.sh
. util.sh

apply-node-taint() {
	local error_count=0 node="" taint_key="" value="" effect=""

	while [[ $# -gt 0 ]]; do
		local key="$1"; shift
		case $key in
		-node)
			node="$1"; shift
			;;
		-key)
			taint_key="$1"; shift
			;;
		-value)
			value="$1"; shift
			;;
		-effect)
			effect="$1"; shift
			;;
		*)
			log.error "$(eval_gettext "error_unrecognized_parameter \$key")"
			return 1
			;;
		esac
	done

	if [[ $node == "" ]]; then
		log.error "$(gettext -s "apply_taint_no_node")"
		error_count=$((error_count + 1))
	fi

	if [[ $taint_key == "" ]]; then
		log.error "$(gettext -s "apply_taint_no_key")"
		error_count=$((error_count + 1))
	fi

	if [[ $value == "" ]]; then
		log.error "$(gettext -s "apply_taint_no_value")"
		error_count=$((error_count + 1))
	fi

	if [[ $effect == "" ]]; then
		log.error "$(gettext -s "apply_taint_no_effect")"
		error_count=$((error_count + 1))
	fi

	case $effect in
	NoSchedule)
		;;
	PreferNoSchedule)
		;;
	NoExecute)
		;;
	*)
		log.error "$(eval_gettext "apply_taint_unrecognized_effect \$effect")"
		error_count=$((error_count + 1))
		;;
	esac

	[[ $error_count -gt 0 ]] && return $error_count

	local -a acceptable=($node)
	local node_fqdn=$(find-best-host acceptable[@])

	if [[ $node_fqdn == "" ]]; then
		log.error "$(gettext -s "apply_taint_node_not_found \$node")"
		error_count=$((error_count + 1))
	fi

	[[ $error_count -gt 0 ]] && return $error_count

	kubectl taint nodes "$node_fqdn" "$taint_key=$value:$effect"

	return $?
}

apply-node-label() {
	local error_count=0 node="" label_key="" value="" effect=""

	while [[ $# -gt 0 ]]; do
		local key="$1"; shift
		case $key in
		-node)
			node="$1"; shift
			;;
		-key)
			label_key="$1"; shift
			;;
		-value)
			value="$1"; shift
			;;
		*)
			log.error "$(eval_gettext "error_unrecognized_parameter \$key")"
			return 1
			;;
		esac
	done

	if [[ $node == "" ]]; then
		log.error "$(gettext -s "apply_label_no_node")"
		error_count=$((error_count + 1))
	fi

	if [[ $label_key == "" ]]; then
		log.error "$(gettext -s "apply_label_no_key")"
		error_count=$((error_count + 1))
	fi

	if [[ $value == "" ]]; then
		log.error "$(gettext -s "apply_label_no_value")"
		error_count=$((error_count + 1))
	fi

	[[ $error_count -gt 0 ]] && return $error_count

	local -a acceptable=($node)
	local node_fqdn=$(find-best-host acceptable[@])

	if [[ $node_fqdn == "" ]]; then
		log.error "$(gettext -s "apply_label_node_not_found \$node")"
		error_count=$((error_count + 1))
	fi

	[[ $error_count -gt 0 ]] && return $error_count

	kubectl label nodes "$node_fqdn" "$label_key=$value"

	return $?
}

wait-for-matching-output() {
	local -r max_time="$1"; shift
	local -r message="$1"; shift
	local -r cmd="$1"; shift
	local -r expected="$1"; shift
	local delay_block_size=15 delay_block=0 last_delay_block=0
	local start_time=$(date +%s) end_time
	local elapsed_time

	echo -n "$(eval_gettext "wait_for \$max_time \$message")"

	actual=$(eval "$cmd") || { echo; log.error "$(eval_gettext "wait_for_cmd_error \$cmd")"; return 1; }
	while [[ "$actual" != "$expected" ]]; do
		sleep 1
		end_time=$(date +%s)
		delay_block=$(( (end_time - start_time) / delay_block_size ))

		if [[ $delay_block != $last_delay_block ]]; then
			echo -n "$(gettext "wait_for_progress")"
			last_delay_block=$delay_block
		fi

		if [[ $((end_time-start_time)) -gt $max_time ]]; then
			elapsed_time=$(convert-seconds $((end_time-start_time)) )
			eval_gettext "wait_for_timeout \$elapsed_time"
			return 1
		fi

		actual=$(eval "$cmd") || { echo; log.error "$(eval_gettext "wait_for_cmd_error \$cmd")"; return 1; }
	done
	end_time=$(date +%s)

	# shellcheck disable=SC2034
	elapsed_time=$(convert-seconds $((end_time-start_time)) )
	eval_gettext "wait_for_done \$elapsed_time"
	echo

	return 0
}

wait-for-all-nodes-ready() {
	local -r expected_nodes="$1"; shift
	local -r message="$(eval_gettext "wait_for_cluster_nodes_ready")"

	wait-for-matching-output 900 "$message" \
		"kubectl get nodes | grep -v NotReady | grep Ready | wc -l" \
		"$expected_nodes"

	return $?
}

wait-for-kube-manager-ready() {
	local rc=0
	local message="$(eval_gettext "wait_for_cluster_manager_ready")"

	log.info "$(eval_gettext "wait_for_cluster_manager_kubectl")"
	# Give the manager a bit of time to get started so kubectl works.
	kubectl get nodes > /dev/null
	while [[ $? -ne 0 ]]; do
		sleep 1
		kubectl get nodes > /dev/null
	done
	log.info "$(eval_gettext "wait_for_cluster_manager_kubectl_ok")"
	sleep 1

	wait-for-matching-output 900 "$message" \
		"kubectl get nodes | grep -v NotReady | grep Ready | wc -l" \
		1
	rc=$?

	if [[ $rc -eq 0 ]]; then
		wait-for-pod-ready "-l k8s-app=kube-dns" "kube-system"
		rc=$?
	fi

	return $rc
}

wait-for-pvc-bound() {
	local -r pvc_name="$1"; shift
	local -r message="$(eval_gettext "wait_for_pvc_bound \$pvc_name")"
	local -r target="$(gettext -s "wait_for_pvc_bound_success_string")"

	wait-for-matching-output 300 "$message" \
		"kubectl get pvc $pvc_name -o yaml | grep phase: | awk '{print \$2}'" \
		"$target"

	return $?
}

wait-for-pod-ready() {
	local -r pod="$1"; shift
	local -r namespace="${1:-default}"; shift
	local -r message="$(eval_gettext "wait_for_pod_ready \$pod")"
	local -r target="$(gettext -s "wait_for_pod_ready_success_string")"
	local pod_name=$(kubectl get pod -n "$namespace" "$pod" --output=jsonpath='{$.items[*].metadata.name}')

	if [[ $pod_name == "" ]]; then
		pod_name="$pod"
	fi

	wait-for-matching-output 900 "$message" \
		"kubectl get pod -n $namespace $pod_name --output=jsonpath='{..containerStatuses[*].ready}' | tr ' ' '\n' | sort -u" \
		"$target"

}

update-nginx-conf() {
        local -r component="$1"
        echo "Enabling the nginx proxy for $component"
        local search_config=$(kubectl get cm fci-common-ui-nginx -o yaml | grep "INCLUDE_CONFIG.*:.*" | grep "[,]*$component\(,\|\$\)")
        if [[ $search_config == "" ]]; then
          SED_REGX="s/\(INCLUDE_CONFIG.*:.*\)\s*$/\1,${component}/"
          kubectl get cm fci-common-ui-nginx -o yaml | sed -e ${SED_REGX} | kubectl replace cm fci-common-ui-nginx -f -
          kubectl delete pods -l app=common-ui-nginx
          wait-for-pod-ready "-l app=common-ui-nginx"
        else
        echo "Proxy already enabled"
        fi
}
