#!/usr/bin/env bash
#
# Licensed Materials - Property of IBM
#
# IBM Financial Crimes Insight with Watson PID 5900-A0H
# IBM Financial Crimes Insight with Watson, Private PID 5737-E41
#
# (C) Copyright IBM Corp. 2017, 2018  All rights reserved.
#
# The source code for this program is not published or otherwise divested
# of its trade secrets, irrespective of what has been deposited with the
# U. S. Copyright Office.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

# Export TEXTDOMAIN before sourcing this file.

# Capture the current working directory
export SCRIPT_DIR=$(cd -P -- "$(dirname -- "$0")" && pwd -P)
export SCRIPT_NAME=$(basename "$0")
export TEXTDOMAINDIR="$SCRIPT_DIR/locale"

. gettext.sh

shopt -s expand_aliases

# If LC_CTYPE is set, it stops the translations from working.
if [[ $LC_CTYPE != "" ]]; then
	unset LC_CTYPE
fi

# Define variables to create a lock on the script
LOCKFILE="${SCRIPT_DIR}/$SCRIPT_NAME.lock"
LOCKFD=99

# Lock functions
_lock()			{ echo $$>"$LOCKFILE"; flock -"$1" "$LOCKFD"; }
_remove_lock()		{ _lock u; _lock xn && rm -f "$LOCKFILE"; }
_init_lock()		{ eval "exec $LOCKFD>\"$LOCKFILE\""; trap _remove_lock EXIT; }
_exclusivelock()	{ _lock xn; }
_alreadylocked()	{ printf "\nUnable to attain an exclusive lock on %s.  Exiting.\nPlease verify that this script isn't already running.\n\n" "$SCRIPT_DIR/${0##*/}"; exit 1; }

# Create lock
_init_lock
_exclusivelock || _alreadylocked

if [[ "$logging_info" == "" ]]; then

# shellcheck disable=SC2034
declare -r logging_info=$(gettext -n -s "log_info")
# shellcheck disable=SC2034
declare -r logging_warning=$(gettext -n -s "log_warning")
# shellcheck disable=SC2034
declare -r logging_error=$(gettext -n -s "log_error")

fi

alias log.info='logging $logging_info $LINENO $INSTALL_STEP; echo '
alias log.warn='logging $logging_warning $LINENO $INSTALL_STEP; echo '
alias log.error='logging $logging_error $LINENO $INSTALL_STEP; echo '
alias log.header='log-header $LINENO $INSTALL_STEP '

logging() {
	# shellcheck disable=SC2034
	local -r log_type="${1}"; shift
	# shellcheck disable=SC2034
	local -r line_no="${1}"; shift
	# shellcheck disable=SC2034
	local -r install_step="${1}"; shift
	# shellcheck disable=SC2034
	local -r date="$(date --rfc-3339=s)"

	eval_gettext "log_record_4 \$date \$log_type \$line_no \$install_step"
}

log-header() {
	local -r line_no="${1}"; shift
	local -r install_step="${1}"; shift
	local -r key="${1}"; shift
	local -r lines="$(gettext -s log_header_lines)"

	logging "$logging_info" "$line_no" "$install_step"
	echo "$lines"

	logging "$logging_info" "$line_no" "$install_step"
	eval_gettext "$key"
	echo

	logging "$logging_info" "$line_no" "$install_step"
	echo "$lines"
}

convert-seconds() {
	local hours minutes seconds

	(( hours   = ${1}/3600 ))
	(( minutes = (${1}%3600)/60 ))
	(( seconds = ${1}%60 ))

	eval_gettext "elapsed_time_3 \$hours \$minutes \$seconds"
}

element-in() {
	local e match="$1"
	shift
	for e; do [[ $e == "$match" ]] && return 0; done
	return 1
}

element-in-array() {
	local match="$1"
	local -a elements=("${!2}")

	for e in "${elements[@]}"; do
		# For some reason, can't quote either of these or
		# worker.* != worker.1
		# shellcheck disable=SC2053
		[[ $match == $e ]] && return 0
	done
	return 1
}

text-in-string() {
	[ -z "${2##*$1*}" ];
}

get-property-value() {
	local -r propertyFile="$1"; shift
	local property="$1"; shift
	local -r default_value="$1"; shift
	local value=""
	local rc=0 

	if [[ "$propertyFile" == "$HOSTNAME_FILE" ]]; then
		property="${property/manager/master}"
	fi

	while true
	do
		[[ "$propertyFile" == "" || "$property" == "" ]] && rc=1 && break 

		property="${property%%[[:space:]]}"		
		property="${property##[[:space:]]}"	

		[[ ${property} == "" ]] && rc=2 && break
		[[ ! -f "${propertyFile}" ]] && rc=3 && break

		value=$(sed '/^\#/d' "${propertyFile}" | grep "^[[:space:]]*${property}[=:[:space:]]" | tail -n 1 )		
		[[ ${value} == ""  && $default_value != "" ]] && value="${property} = ${default_value}"
		[[ ${value} == "" ]] && rc=4 && break

		# Remove leading/trailing white space 
		value="${value%%[[:space:]]}"		
		value="${value##[[:space:]]}"	
		# Isolate value (split on ':' or '=')
		value=$(echo "${value}" | sed 's/[^:=]*//')
		# delete everything after the first # (comment)
		value=${value%%"#"*}
		# Remove leading value whitespace
		value=$(echo "${value}" | sed 's/^[:=][[:space:]]*//')
		# Remove "Windows" CR character (if any) from value.
		value=$(echo "${value}" | sed 's|\r$||')
		break
	done
	echo -n "${value}"
	return $rc
}

is_install_step_done() {
	local -r install_step="${1}"; shift
	local -r lock_file="${1}"; shift
	local rc=0

	if [[ -f "$lock_file" ]]; then
		# shellcheck disable=SC2034
		local -r note="$(cat "$lock_file")"
		log.info "$(eval_gettext "log_step_already_completed_2 \$install_step \$note")"
		rc=1
	fi
	return $rc
}

install_step_done() {
	local -r lock_file="${1}"; shift
	# shellcheck disable=SC2034
	local -r when=$(date)

	eval_gettext "log_step_completion_note_1 \$when" > "$lock_file"
}

check-property-file() {
	# shellcheck disable=SC2034
	local -r directory="${1}"; shift
	local -r filename="${1}"; shift
	local rc=0

	if [[ ! -f $filename ]]; then
		log.error "$(eval_gettext "property_file_not_found \$directory \$filename")"
		rc=1
	else
		if ! rpm -q dos2unix; then
			if ! yum -y install dos2unix; then
				return 1
			fi
		fi
		dos2unix "$filename"
		log.info "$(eval_gettext "property_file_found \$directory \$filename")"
	fi

	return $rc
}

getSoftwarePackageVersion() {
	local -r package="$1"; shift
	local -r my_hostname="$1"; shift
	local -r HOSTNAME_FQDN="$1"; shift

	if [[ "$my_hostname" != "$HOSTNAME_FQDN" ]]; then
				# shellcheck disable=SC2029
		ssh -n "$HOSTNAME_FQDN" "rpm -q --qf \"${VERSION}\" $package"
	else
		rpm -q --qf "%{VERSION}" "$package"
	fi
}
