#!/usr/bin/env bash

declare dns_ip="$(grep -- --cluster-dns= /etc/systemd/system/kubelet.service.d/10-kubeadm.conf | tr '-' '\n' | grep dns | awk -F= '{print $2}')"
declare calico_etcd_ip="$(kubectl describe service -n kube-system calico-etcd | grep '^IP:' | awk '{print $2}')"

convert-seconds() {
        local hours minutes seconds

        (( hours   = ${1}/3600 ))
        (( minutes = (${1}%3600)/60 ))
        (( seconds = ${1}%60 ))

        echo "[$hours hours $minutes minutes $seconds seconds elapsed time]"
}


wait-for-matching-output() {
        local -r max_time="$1"; shift
        local -r message="$1"; shift
        local -r cmd="$1"; shift
        local -r expected="$1"; shift
        local delay_block_size=15 delay_block=0 last_delay_block=0
        local start_time=$(date +%s) end_time
        local elapsed_time

        echo -n "Waiting up to $max_time seconds for $message:  "

        actual=$(eval "$cmd") || { echo; echo "Failed to execute command:  $cmd"; return 1; }
        while [[ "$actual" != "$expected" ]]; do
                sleep 1
                end_time=$(date +%s)
                delay_block=$(( (end_time - start_time) / delay_block_size ))

                if [[ $delay_block != $last_delay_block ]]; then
                        echo -n "."
                        last_delay_block=$delay_block
                fi

                if [[ $((end_time-start_time)) -gt $max_time ]]; then
                        elapsed_time=$(convert-seconds $((end_time-start_time)) )
                        echo "Timed out $elapsed_time"
                        return 1
                fi

                actual=$(eval "$cmd") || { echo; echo "Failed to execute command:  $cmd"; return 1; }
        done
        end_time=$(date +%s)

        # shellcheck disable=SC2034
        elapsed_time=$(convert-seconds $((end_time-start_time)) )
        echo "OK $elapsed_time"
        echo

        return 0
}


wait-for-all-nodes-ready() {
	local -r debugging="$( [[ "$-" =~ .*x.* ]] && echo 1 || echo 0 )"
	set +x
	local -r expected_nodes="$(kubectl get nodes | grep -v NAME  -c)"
        local -r message="all nodes in the cluster to be ready"

        wait-for-matching-output 1800 "$message" \
                "kubectl get nodes | grep -v NotReady | grep Ready | wc -l" \
                "$expected_nodes"

	if [[ "$debugging" == "1" ]]; then
		set -x
	fi

        return $?
}

wait-for-all-pods-ready() {
	local -r debugging="$( [[ "$-" =~ .*x.* ]] && echo 1 || echo 0 )"
	set +x

        local -r message="all pods to be ready"
        local -r target="true"

        wait-for-matching-output 1800 "$message" \
                "kubectl get pod --all-namespaces -l !job-name --output=jsonpath='{..containerStatuses[*].ready}' | tr ' ' '\n' | sort -u" \
                "$target" || return $?

	if [[ "$debugging" == "1" ]]; then
		set -x
	fi
	return 0
}

install_rpms() {
	local -r version="${1}"; shift
	local -r target="${1}"; shift
	local -r dir="tmp-apar-$$"

	# shellcheck disable=SC2029
	ssh "$target" "mkdir $dir"
	scp ./*.rpm "$target:$dir/"

	local CNI_RPM=""

	if [[ "$version" =~ 1.9.* || "$version" == "1.10.11" ]]; then
		CNI_RPM+="*kubernetes-cni-0.6.0* "
		CNI_RPM+="*cri-tools-1.11.1* "
	fi

	local cmd="cd $dir && "
	cmd+="yum -y install --nogpgcheck --disablerepo='*' *${version}*.rpm $CNI_RPM; "
	cmd+="cd .. && rm -fr $dir"

	# shellcheck disable=SC2029
	ssh "$target" "$cmd"
}

load_docker_images() {
	local -r repo="${1}"; shift
	local -r version="${1}"; shift
	local -a image_tars=()

	case "$version" in
	1.8.14)
		image_tars=(kube-apiserver-amd64-1.8.14.tar \
			kube-controller-manager-amd64-1.8.14.tar \
			kube-scheduler-amd64-1.8.14.tar \
			kube-proxy-amd64-1.8.14.tar \
			etcd-amd64-3.0.17.tar \
			pause-amd64-3.0.tar \
			k8s-dns-sidecar-amd64-1.14.5.tar \
			k8s-dns-kube-dns-amd64-1.14.5.tar \
			k8s-dns-dnsmasq-nanny-amd64-1.14.5.tar \
			centos-7.tar \
			defaultbackend-1.0.tar \
			nginx-ingress-controller-0.9.0-beta.11.tar \
			calico-cni-v2.0.0.tar \
			calico-kube-controllers-v2.0.0.tar \
			calico-node-v3.0.2.tar \
			coreos-etcd-v3.1.10.tar \
		)
		;;
	1.9.11)
		image_tars=(kube-apiserver-amd64-1.9.11.tar \
			kube-controller-manager-amd64-1.9.11.tar \
			kube-scheduler-amd64-1.9.11.tar \
			kube-proxy-amd64-1.9.11.tar \
			etcd-amd64-3.1.11.tar \
			k8s-dns-sidecar-amd64-1.14.7.tar \
			k8s-dns-kube-dns-amd64-1.14.7.tar \
			k8s-dns-dnsmasq-nanny-amd64-1.14.7.tar \
		)
		;;
	1.10.11)
		image_tars=(kube-apiserver-amd64-1.10.11.tar \
			kube-controller-manager-amd64-1.10.11.tar \
			kube-scheduler-amd64-1.10.11.tar \
			kube-proxy-amd64-1.10.11.tar \
			etcd-amd64-3.1.12.tar \
			k8s-dns-sidecar-amd64-1.14.8.tar \
			k8s-dns-kube-dns-amd64-1.14.8.tar \
			k8s-dns-dnsmasq-nanny-amd64-1.14.8.tar \
			pause-amd64-3.1.tar \
		)
		;;
	esac

	for image_tar in "${image_tars[@]}"; do
		local image="$(docker load < "${image_tar}" | awk '{print $3}')"
		[[ $? == "0" ]] || exit 1

		local dest="${image#*/}"
		docker tag "${image}" "${repo}/${dest}" && \
		docker push "${repo}/${dest}"
		[[ $? == "0" ]] || exit 1
	done
}

upgrade_kube_cluster() {
	local -r target_ver="${1}"; shift
	local -r svr_ver="$(kubectl version --short | grep Server | awk '{print $3}')"
	if [[ "v$target_ver" != "$svr_ver" ]]; then
		if ! kubeadm upgrade apply -y "v${target_ver}"; then
			wait-for-all-nodes-ready && \
			wait-for-all-pods-ready && \
			sleep 60 || return $?
			if ! kubeadm upgrade apply -y "v${target_ver}"; then
				wait-for-all-nodes-ready && \
				wait-for-all-pods-ready && \
				sleep 60 && \
				kubeadm upgrade apply -y "v${target_ver}" || \
				return $?
			fi
		fi
	fi
	wait-for-all-nodes-ready && \
	wait-for-all-pods-ready || \
	return $?
}

rationalize_cgroup() {
	local -r server="${1}"; shift
	local -r DOCKER_CMD_CGROUP=$(ssh "$server" "docker info" 2>&1 | grep -i cgroup)
	local -r DOCKER_PARTIAL_CGROUP="${DOCKER_CMD_CGROUP//[^[:alpha:]^:]/}"
	local -r DOCKER_CGROUP="${DOCKER_PARTIAL_CGROUP##*:}"
	local -r KUBE_FILE=/etc/systemd/system/kubelet.service.d/10-kubeadm.conf

### TODO
### merge these two ssh commands into one

	# shellcheck disable=SC2029
	ssh "$server" "if ! grep -q \"cgroup-driver=${DOCKER_CGROUP}\" $KUBE_FILE; then sed -i \"s/cgroup-driver=.*\\\"/cgroup-driver=${DOCKER_CGROUP}\\\"/\" $KUBE_FILE; fi"
	# shellcheck disable=SC2029
	ssh "$server" "sed -i 's/--cluster-dns=10.96.0.10/--cluster-dns=${dns_ip}/' /etc/systemd/system/kubelet.service.d/10-kubeadm.conf"
}

upgrade_kubelet() {
	local -r server="${1}"; shift

	local -r DOCKER_CMD_CGROUP=$(ssh "$server" "docker info" 2>&1 | grep -i cgroup)
	local -r DOCKER_PARTIAL_CGROUP="${DOCKER_CMD_CGROUP//[^[:alpha:]^:]/}"
	local -r DOCKER_CGROUP="${DOCKER_PARTIAL_CGROUP##*:}"
	local -r KUBE_FILE=/etc/systemd/system/kubelet.service.d/10-kubeadm.conf

	kubectl drain "$server" --force --delete-local-data --ignore-daemonsets || exit $?

	local cmd=""

	cmd+="if ! grep -q \"cgroup-driver=${DOCKER_CGROUP}\" $KUBE_FILE; then "
	cmd+="sed -i \"s/cgroup-driver=.*\\\"/cgroup-driver=${DOCKER_CGROUP}\\\"/\" "
	cmd+="       $KUBE_FILE; "
	cmd+="fi; "

	## On !master nodes:
	#if [[ "$server" != "$(hostname -f)" ]]; then
	#	cmd+="kubeadm upgrade node config --kubelet-version"
	#	cmd+=" \$(kubelet --version | cut -d\\   -f 2); "
	#fi

	cmd+="systemctl daemon-reload && "
	cmd+="systemctl restart kubelet && "
	cmd+="sleep 30"

	# shellcheck disable=SC2029
	ssh "$server" "$cmd" || exit $?
	kubectl uncordon "$server" || exit $?
	wait-for-all-nodes-ready && \
	wait-for-all-pods-ready || \
	return $?
}

convert_to_offline_install() {
	local -r docker_repo="$(kubectl describe pods | grep Image: | grep ibmcom | awk '{print $2}' | awk -F/ '{print $1}' | sort -u)"

	if kubectl get ds -n kube-system calico-etcd -o yaml | grep 'quay.io'; then 
		kubectl get deploy fci-analytics -o yaml | sed -e "s|image: centos:7|image: ${docker_repo}/centos:7|g" | kubectl apply -f - && \
		kubectl get deploy fci-solution -o yaml | sed -e "s|image: centos:7|image: ${docker_repo}/centos:7|g" | kubectl apply -f - && \
		kubectl get deploy -n kube-system default-http-backend -o yaml | sed -e "s|image: gcr.io|image: ${docker_repo}|g" | kubectl apply -f - && \
		kubectl get deploy -n kube-system nginx-ingress-controller -o yaml | sed -e "s|image: gcr.io|image: ${docker_repo}|g" | kubectl apply -f - && \
		kubectl get ds -n kube-system calico-etcd -o yaml | sed -e "s|image: quay.io|image: ${docker_repo}|g" | kubectl apply -f - && \
		kubectl get ds -n kube-system calico-node -o yaml | sed -e "s|image: quay.io|image: ${docker_repo}|g" | kubectl apply -f - && \
		kubectl get deploy -n kube-system calico-kube-controllers -o yaml | sed -e "s|image: quay.io|image: ${docker_repo}|g" | kubectl apply -f - && \
		wait-for-all-pods-ready || \
		return $?
	fi
	return 0
}

upgrade_to_kube_version() {
	local -r target_version="${1}"; shift
	local load_docker_repo=1; shift

	for node in $all_nodes; do
		install_rpms "$target_version" "$node" && \
		rationalize_cgroup "$node" || exit 1
	done

	if ! kubeadm config view > /dev/null ; then
		local advertise_addr=$(grep -- -advertise-address /etc/kubernetes/manifests/kube-apiserver.yaml | awk -F= '{print $2}')
		local pod_cidr=$(grep -- -cluster-cidr /etc/kubernetes/manifests/kube-controller-manager.yaml | awk -F= '{print $2}')
		local svc_cidr=$(grep -- -service-cluster-ip-range /etc/kubernetes/manifests/kube-apiserver.yaml | awk -F= '{print $2}')

		kubeadm config upload from-flags \
			--apiserver-cert-extra-sans "$advertise_addr" \
			--pod-network-cidr="${pod_cidr}" \
			--service-cidr="${svc_cidr}"
	fi

	local -r docker_repo="$(kubectl describe pods | grep Image: | grep ibmcom | awk '{print $2}' | awk -F/ '{print $1}' | sort -u)"

	load_docker_images "$docker_repo" "$target_version" || exit $?

	upgrade_kube_cluster "$target_version"

	if [[ "$target_version" == "1.8.14" ]]; then
		if ! kubectl -n kube-system get cm kubeadm-config -o yaml | \
			grep -q "$docker_repo"; then
			# online install, let's fix that.
			kubectl -n kube-system get cm kubeadm-config -o yaml | \
				sed "s|imageRepository:.*|imageRepository: $docker_repo/google_containers|" | \
				kubectl apply -f - || exit $?
		fi
	fi

	local -r all_nodes="$(kubectl get nodes | awk '{print $1}' | grep -v NAME)"
	for node in $all_nodes; do
		upgrade_kubelet "$node"
	done
}

main() {
	local -r all_nodes="$(kubectl get nodes | awk '{print $1}' | grep -v NAME)"
	local load_docker_repo=1
	local svr_ver

	wait-for-all-nodes-ready && \
	wait-for-all-pods-ready || \
	return $?

	svr_ver="$(kubectl version --short | grep Server | awk '{print $3}')"

	while [[ "$svr_ver" =~ v1.7.* ]]; do
		upgrade_to_kube_version 1.8.14 "${load_docker_repo}" && \
		convert_to_offline_install || exit $?
		svr_ver="$(kubectl version --short | grep Server | awk '{print $3}')"
	done

	while [[ "$svr_ver" =~ v1.8.* ]]; do
		upgrade_to_kube_version 1.9.11 "${load_docker_repo}"
		svr_ver="$(kubectl version --short | grep Server | awk '{print $3}')"
	done

	while [[ "$svr_ver" =~ v1.9.* || "$svr_ver" == "v1.10.1" ]]; do
		upgrade_to_kube_version 1.10.11 "${load_docker_repo}"
		svr_ver="$(kubectl version --short | grep Server | awk '{print $3}')"
	done

}

main "$(basename "$0")" "$@"
